@@ -3,7 +3,7 @@
  * Log-structured Caching for Linux
  *
  * This file is part of dm-writeboost
- * Copyright (C) 2012-2024 Akira Hayakawa <ruby.wktk@gmail.com>
+ * Copyright (C) 2012-2025 Akira Hayakawa <ruby.wktk@gmail.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -708,7 +708,7 @@ struct per_bio_data {
 		struct segment_header *seg;
 	};
 };
-#define per_bio_data(wb, bio) ((struct per_bio_data *)dm_per_bio_data((bio), (wb)->ti->PER_BIO_DATA_SIZE))
+#define per_bio_data(bio) ((struct per_bio_data *) dm_per_bio_data((bio), sizeof(struct per_bio_data)))
 
 /*----------------------------------------------------------------------------*/
 
@@ -732,10 +732,10 @@ static void read_cache_add(struct read_cache_cells *cells, struct read_cache_cel
 	rb_insert_color(&cell->rb_node, &cells->rb_root);
 }
 
-static struct read_cache_cell *lookup_read_cache_cell(struct wb_device *wb, sector_t sector)
+static struct read_cache_cell *lookup_read_cache_cell(struct read_cache_cells *cells, sector_t sector)
 {
 	struct rb_node **rbp, *parent;
-	rbp = &wb->read_cache_cells->rb_root.rb_node;
+	rbp = &cells->rb_root.rb_node;
 	parent = NULL;
 	while (*rbp) {
 		struct read_cache_cell *parent_cell;
@@ -752,7 +752,10 @@ static struct read_cache_cell *lookup_read_cache_cell(struct wb_device *wb, sect
 	return NULL;
 }
 
-static void read_cache_cancel_cells(struct read_cache_cells *cells, u32 n)
+/*
+ * Cancel all cells in [cursor, cursor + seqcount).
+ */
+static void read_cache_cancel_seq_cells(struct read_cache_cells *cells)
 {
 	u32 i;
 	u32 last = cells->cursor + cells->seqcount;
@@ -760,7 +763,7 @@ static void read_cache_cancel_cells(struct read_cache_cells *cells, u32 n)
 		last = cells->size;
 	for (i = cells->cursor; i < last; i++) {
 		struct read_cache_cell *cell = cells->array + i;
-		cell->cancelled = true;
+		atomic_set(&cell->cancelled, 1);
 	}
 }
 
@@ -777,32 +780,54 @@ static void read_cache_cancel_foreground(struct read_cache_cells *cells,
 		cells->seqcount = 1;
 		cells->over_threshold = false;
 	}
+	cells->last_sector = new_cell->sector;
 
 	if (cells->seqcount > cells->threshold) {
 		if (cells->over_threshold)
-			new_cell->cancelled = true;
+			atomic_set(&new_cell->cancelled, 1);
 		else {
 			cells->over_threshold = true;
-			read_cache_cancel_cells(cells, cells->seqcount);
+			read_cache_cancel_seq_cells(cells);
 		}
 	}
-	cells->last_sector = new_cell->sector;
+}
+
+static bool do_reserve_read_cache_cell(struct read_cache_cells *cells, struct bio *bio)
+{
+	struct read_cache_cell *found, *new_cell;
+
+	ASSERT(cells->threshold > 0);
+	if (!cells->cursor)
+		return false;
+
+	/*
+	 * We don't need to reserve the same address twice
+	 * because it's either unchanged or invalidated.
+	 */
+	found = lookup_read_cache_cell(cells, bi_sector(bio));
+	if (found)
+		return false;
+
+	cells->cursor--;
+	new_cell = cells->array + cells->cursor;
+	new_cell->sector = bi_sector(bio);
+	read_cache_add(cells, new_cell);
+
+	/* Cancel the new_cell if needed */
+	read_cache_cancel_foreground(cells, new_cell);
+
+	return true;
 }
 
 static bool reserve_read_cache_cell(struct wb_device *wb, struct bio *bio)
 {
 	struct per_bio_data *pbd;
 	struct read_cache_cells *cells = wb->read_cache_cells;
-	struct read_cache_cell *found, *new_cell;
-
-	ASSERT(cells->threshold > 0);
+	bool reserved;
 
 	if (!read_once(wb->read_cache_threshold))
 		return false;
 
-	if (!cells->cursor)
-		return false;
-
 	/*
 	 * We only cache 4KB read data for following reasons:
 	 * 1) Caching partial data (< 4KB) is likely meaningless.
@@ -811,40 +836,36 @@ static bool reserve_read_cache_cell(struct wb_device *wb, struct bio *bio)
 	if (!bio_is_fullsize(bio))
 		return false;
 
-	/*
-	 * We don't need to reserve the same address twice
-	 * because it's either unchanged or invalidated.
-	 */
-	found = lookup_read_cache_cell(wb, bi_sector(bio));
-	if (found)
+	mutex_lock(&cells->lock);
+	reserved = do_reserve_read_cache_cell(cells, bio);
+	mutex_unlock(&cells->lock);
+
+	if (!reserved)
 		return false;
 
-	cells->cursor--;
-	new_cell = cells->array + cells->cursor;
-	new_cell->sector = bi_sector(bio);
-	read_cache_add(cells, new_cell);
-
-	pbd = per_bio_data(wb, bio);
+	pbd = per_bio_data(bio);
 	pbd->type = PBD_WILL_CACHE;
 	pbd->cell_idx = cells->cursor;
 
-	/* Cancel the new_cell if needed */
-	read_cache_cancel_foreground(cells, new_cell);
-
 	return true;
 }
 
 static void might_cancel_read_cache_cell(struct wb_device *wb, struct bio *bio)
 {
 	struct read_cache_cell *found;
-	found = lookup_read_cache_cell(wb, calc_cache_alignment(bi_sector(bio)));
+	struct read_cache_cells *cells = wb->read_cache_cells;
+
+	mutex_lock(&cells->lock);
+	found = lookup_read_cache_cell(cells, calc_cache_alignment(bi_sector(bio)));
+	mutex_unlock(&cells->lock);
+
 	if (found)
-		found->cancelled = true;
+		atomic_set(&found->cancelled, 1);
 }
 
 static void read_cache_cell_copy_data(struct wb_device *wb, struct bio *bio, unsigned long error)
 {
-	struct per_bio_data *pbd = per_bio_data(wb, bio);
+	struct per_bio_data *pbd = per_bio_data(bio);
 	struct read_cache_cells *cells = wb->read_cache_cells;
 	struct read_cache_cell *cell = cells->array + pbd->cell_idx;
 
@@ -852,13 +873,13 @@ static void read_cache_cell_copy_data(struct wb_device *wb, struct bio *bio, uns
 
 	/* Data can be broken. So don't stage. */
 	if (error)
-		cell->cancelled = true;
+		atomic_set(&cell->cancelled, 1);
 
 	/*
 	 * We can omit copying if the cell is cancelled but
 	 * copying for a non-cancelled cell isn't problematic.
 	 */
-	if (!cell->cancelled)
+	if (!atomic_read(&cell->cancelled))
 		copy_bio_payload(cell->data, bio);
 
 	if (atomic_dec_and_test(&cells->ack_count))
@@ -884,7 +905,7 @@ static void inject_read_cache(struct wb_device *wb, struct read_cache_cell *cell
 	 * if might_cancel_read_cache_cell() on the foreground
 	 * cancelled this cell, the data is now stale.
 	 */
-	if (cell->cancelled) {
+	if (atomic_read(&cell->cancelled)) {
 		mutex_unlock(&wb->io_lock);
 		return;
 	}
@@ -929,6 +950,7 @@ static struct read_cache_cells *alloc_read_cache_cells(struct wb_device *wb, u32
 	if (!cells)
 		return NULL;
 
+	mutex_init(&cells->lock);
 	cells->size = n;
 	cells->threshold = UINT_MAX; /* Default: every read will be cached */
 	cells->last_sector = ~0;
@@ -975,25 +997,21 @@ static void free_read_cache_cells(struct wb_device *wb)
 	kfree(cells);
 }
 
-static void reinit_read_cache_cells(struct wb_device *wb)
+static void reinit_read_cache_cells(struct read_cache_cells *cells, u32 new_threshold)
 {
-	struct read_cache_cells *cells = wb->read_cache_cells;
-	u32 i, cur_threshold;
+	u32 i;
 
-	mutex_lock(&wb->io_lock);
 	cells->rb_root = RB_ROOT;
 	cells->cursor = cells->size;
 	atomic_set(&cells->ack_count, cells->size);
 	for (i = 0; i < cells->size; i++) {
 		struct read_cache_cell *cell = cells->array + i;
-		cell->cancelled = false;
+		atomic_set(&cell->cancelled, 0);
 	}
-	cur_threshold = read_once(wb->read_cache_threshold);
-	if (cur_threshold && (cur_threshold != cells->threshold)) {
-		cells->threshold = cur_threshold;
+	if (new_threshold && (new_threshold != cells->threshold)) {
+		cells->threshold = new_threshold;
 		cells->over_threshold = false;
 	}
-	mutex_unlock(&wb->io_lock);
 }
 
 /*
@@ -1004,7 +1022,7 @@ static void visit_and_cancel_cells(struct rb_node *first, struct rb_node *last)
 	struct rb_node *rbp = first;
 	while (rbp != last) {
 		struct read_cache_cell *cell = read_cache_cell_from_node(rbp);
-		cell->cancelled = true;
+		atomic_set(&cell->cancelled, 1);
 		rbp = rb_next(rbp);
 	}
 }
@@ -1049,7 +1067,9 @@ static void read_cache_proc(struct work_struct *work)
 		inject_read_cache(wb, cell);
 	}
 
-	reinit_read_cache_cells(wb);
+	mutex_lock(&cells->lock);
+	reinit_read_cache_cells(cells, read_once(wb->read_cache_threshold));
+	mutex_unlock(&cells->lock);
 }
 
 static int init_read_cache_cells(struct wb_device *wb)
@@ -1060,7 +1080,7 @@ static int init_read_cache_cells(struct wb_device *wb)
 	if (!cells)
 		return -ENOMEM;
 	wb->read_cache_cells = cells;
-	reinit_read_cache_cells(wb);
+	reinit_read_cache_cells(cells, wb->read_cache_threshold);
 	return 0;
 }
 
@@ -1160,7 +1180,6 @@ static int do_process_write(struct wb_device *wb, struct bio *bio)
 	mutex_lock(&wb->io_lock);
 
 	cache_lookup(wb, bio, &res);
-
 	if (res.found) {
 		if (unlikely(res.on_buffer)) {
 			write_pos = res.found_mb;
@@ -1171,8 +1190,8 @@ static int do_process_write(struct wb_device *wb, struct bio *bio)
 			if (err)
 				goto out;
 		}
-	} else
-		might_cancel_read_cache_cell(wb, bio);
+	}
+	might_cancel_read_cache_cell(wb, bio);
 
 	might_queue_current_buffer(wb);
 
@@ -1261,7 +1280,6 @@ static int process_write_wa(struct wb_device *wb, struct bio *bio)
 		dec_inflight_ios(wb, res.found_seg);
 		ht_del(wb, res.found_mb);
 	}
-
 	might_cancel_read_cache_cell(wb, bio);
 	mutex_unlock(&wb->io_lock);
 
@@ -1428,7 +1446,7 @@ read_mb_exit:
 		return DM_MAPIO_SUBMITTED;
 	}
 
-	pbd = per_bio_data(wb, bio);
+	pbd = per_bio_data(bio);
 	pbd->type = PBD_READ_SEG;
 	pbd->seg = res.found_seg;
 
@@ -1456,7 +1474,7 @@ static int writeboost_map(struct dm_target *ti, struct bio *bio)
 {
 	struct wb_device *wb = ti->private;
 
-	struct per_bio_data *pbd = per_bio_data(wb, bio);
+	struct per_bio_data *pbd = per_bio_data(bio);
 	pbd->type = PBD_NONE;
 
 	if (bio_is_barrier(bio))
@@ -1480,7 +1498,7 @@ static int writeboost_end_io(struct dm_target *ti, struct bio *bio, int error)
 #endif
 {
 	struct wb_device *wb = ti->private;
-	struct per_bio_data *pbd = per_bio_data(wb, bio);
+	struct per_bio_data *pbd = per_bio_data(bio);
 
 	switch (pbd->type) {
 	case PBD_NONE:
@@ -1900,7 +1918,11 @@ static int writeboost_iterate_devices(struct dm_target *ti,
 
 static void writeboost_io_hints(struct dm_target *ti, struct queue_limits *limits)
 {
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6,12,0)
+	limits->io_opt = 4096;
+#else
 	blk_limits_io_opt(limits, 4096);
+#endif
 }
 
 static void writeboost_status(struct dm_target *ti, status_type_t type,
@@ -1968,7 +1990,7 @@ static void writeboost_status(struct dm_target *ti, status_type_t type,
 
 static struct target_type writeboost_target = {
 	.name = "writeboost",
-	.version = {2, 2, 18},
+	.version = {2, 2, 19},
 	.module = THIS_MODULE,
 	.map = writeboost_map,
 	.end_io = writeboost_end_io,
