@@ -1,6 +1,6 @@
 /*
  * This file is part of dm-writeboost
- * Copyright (C) 2012-2024 Akira Hayakawa <ruby.wktk@gmail.com>
+ * Copyright (C) 2012-2025 Akira Hayakawa <ruby.wktk@gmail.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -207,11 +207,13 @@ struct writeback_segment {
 struct read_cache_cell {
 	sector_t sector;
 	void *data; /* 4KB data read */
-	bool cancelled; /* Don't include this */
+	atomic_t cancelled; /* Don't include this */
 	struct rb_node rb_node;
 };
 
 struct read_cache_cells {
+	struct mutex lock;
+
 	u32 size;
 	struct read_cache_cell *array;
 	u32 cursor;
